@@ -1,5 +1,4 @@
-"""ViSP science calibration task."""
-import logging
+"""ViSP science calibration task. See :doc:`this page </science_calibration>` for more information."""
 from collections import defaultdict
 from dataclasses import dataclass
 from functools import cached_property
@@ -9,21 +8,26 @@ import numpy as np
 from astropy.io import fits
 from astropy.time import Time
 from astropy.time import TimeDelta
+from dkist_processing_common.codecs.fits import fits_access_decoder
+from dkist_processing_common.codecs.fits import fits_array_decoder
+from dkist_processing_common.codecs.fits import fits_hdulist_encoder
+from dkist_processing_common.models.task_name import TaskName
 from dkist_processing_common.tasks.mixin.quality import QualityMixin
 from dkist_processing_math.arithmetic import divide_arrays_by_array
 from dkist_processing_math.arithmetic import subtract_array_from_arrays
+from dkist_processing_math.linear_algebra import nd_left_matrix_multiply
 from dkist_processing_math.statistics import average_numpy_arrays
 from dkist_processing_pac.optics.telescope import Telescope
+from dkist_service_configuration.logging import logger
 
 from dkist_processing_visp.models.tags import VispTag
 from dkist_processing_visp.parsers.visp_l0_fits_access import VispL0FitsAccess
+from dkist_processing_visp.tasks.mixin.beam_access import BeamAccessMixin
 from dkist_processing_visp.tasks.mixin.corrections import CorrectionsMixin
-from dkist_processing_visp.tasks.mixin.input_frame_loaders import InputFrameLoadersMixin
-from dkist_processing_visp.tasks.mixin.intermediate_frame_helpers import (
-    IntermediateFrameHelpersMixin,
-)
 from dkist_processing_visp.tasks.visp_base import VispTaskBase
 
+__all__ = ["ScienceCalibration"]
+
 
 @dataclass
 class CalibrationCollection:
@@ -31,7 +35,6 @@ class CalibrationCollection:
 
     dark: dict
     background: dict
-    lamp_gain: dict
     solar_gain: dict
     angle: dict
     state_offset: dict
@@ -46,7 +49,7 @@ class CalibrationCollection:
         This is done by considering that state offset values computed by the GeometricCalibration task. Any sub-pixel
         overlaps are rounded to the next integer that still guarantees overlap.
         """
-        logging.info("Computing beam overlap slices")
+        logger.info("Computing beam overlap slices")
         # This will be a flat list of (x, y) pairs for all modstates and beams
         flat_offsets = sum([list(i.values()) for i in self.state_offset.values()], [])
         # Split out into an x list and a y list
@@ -54,8 +57,8 @@ class CalibrationCollection:
         all_x_shifts = np.array(all_x_shifts)
         all_y_shifts = np.array(all_y_shifts)
 
-        logging.info(f"All x shifts: {all_x_shifts}")
-        logging.info(f"All y shifts: {all_y_shifts}")
+        logger.info(f"All x shifts: {all_x_shifts}")
+        logger.info(f"All y shifts: {all_y_shifts}")
 
         # The amount we need to "slice in" from the front of the array is simply the maximum positive shift
         # The call to `np.ceil` ensures that the integer rounding doesn't allow non-overlap regions to leak in.
@@ -82,8 +85,7 @@ class CalibrationCollection:
 
 class ScienceCalibration(
     VispTaskBase,
-    IntermediateFrameHelpersMixin,
-    InputFrameLoadersMixin,
+    BeamAccessMixin,
     CorrectionsMixin,
     QualityMixin,
 ):
@@ -94,10 +96,17 @@ class ScienceCalibration(
     ----------
     recipe_run_id : int
         id of the recipe run used to identify the workflow run this task is part of
+
     workflow_name : str
         name of the workflow to which this instance of the task belongs
+
     workflow_version : str
         version of the workflow to which this instance of the task belongs
+
+    Returns
+    -------
+    None
+
     """
 
     record_provenance = True
@@ -131,12 +140,12 @@ class ScienceCalibration(
                 tags=[
                     VispTag.input(),
                     VispTag.frame(),
-                    VispTag.task("OBSERVE"),
+                    VispTag.task_observe(),
                 ],
             )
 
             self.quality_store_task_type_counts(
-                task_type="OBSERVE", total_frames=no_of_raw_science_frames
+                task_type=TaskName.observe.value, total_frames=no_of_raw_science_frames
             )
 
     def collect_calibration_objects(self) -> CalibrationCollection:
@@ -147,7 +156,6 @@ class ScienceCalibration(
         """
         dark_dict = defaultdict(dict)
         background_dict = dict()
-        lamp_dict = defaultdict(dict)
         solar_dict = defaultdict(dict)
         angle_dict = dict()
         state_offset_dict = defaultdict(dict)
@@ -155,73 +163,95 @@ class ScienceCalibration(
         demod_dict = dict() if self.constants.correct_for_polarization else None
 
         for beam in range(1, self.constants.num_beams + 1):
-            for exp_time in self.constants.observe_exposure_times:
+            for readout_exp_time in self.constants.observe_readout_exp_times:
                 # Dark
                 ######
-                try:
-                    dark_array = self.intermediate_frame_helpers_load_dark_array(
-                        beam=beam, exposure_time=exp_time
+                dark_array = next(
+                    self.read(
+                        tags=VispTag.intermediate_frame_dark(
+                            beam=beam, readout_exp_time=readout_exp_time
+                        ),
+                        decoder=fits_array_decoder,
                     )
-                except StopIteration:
-                    raise ValueError(f"No matching dark found for {exp_time = } s")
+                )
 
-                dark_dict[VispTag.beam(beam)][VispTag.exposure_time(exp_time)] = dark_array
+                dark_dict[VispTag.beam(beam)][
+                    VispTag.readout_exp_time(readout_exp_time)
+                ] = dark_array
 
             # Residual background light
             ###########################
-            try:
-                background_dict[
-                    VispTag.beam(beam)
-                ] = self.intermediate_frame_helpers_load_background_array(beam=beam)
-            except StopIteration:
-                if self.constants.correct_for_polarization:
-                    raise ValueError(f"No matching background light found for {beam = }")
-                background_dict[VispTag.beam(beam)] = np.zeros(dark_array.shape)
+            background_dict[VispTag.beam(beam)] = np.zeros(dark_array.shape)
+            if self.constants.correct_for_polarization and self.parameters.background_on:
+                background_generator = self.read(
+                    tags=[VispTag.intermediate_frame(beam=beam), VispTag.task_background()],
+                    decoder=fits_array_decoder,
+                )
+                background_dict[VispTag.beam(beam)] = next(background_generator)
 
             # Angle
             #######
-            angle_dict[VispTag.beam(beam)] = self.intermediate_frame_helpers_load_angle(beam=beam)
+            angle_array = next(
+                self.read(
+                    tags=[VispTag.intermediate_frame(beam=beam), VispTag.task_geometric_angle()],
+                    decoder=fits_array_decoder,
+                )
+            )
+            angle_dict[VispTag.beam(beam)] = angle_array[0]
 
             # Spec shifts
             #############
-            spec_shift_dict[VispTag.beam(beam)] = self.intermediate_frame_helpers_load_spec_shift(
-                beam=beam
+            spec_shift_dict[VispTag.beam(beam)] = next(
+                self.read(
+                    tags=[
+                        VispTag.intermediate_frame(beam=beam),
+                        VispTag.task_geometric_spectral_shifts(),
+                    ],
+                    decoder=fits_array_decoder,
+                )
             )
 
             # Demod
             #######
             if self.constants.correct_for_polarization:
-                demod_dict[
-                    VispTag.beam(beam)
-                ] = self.intermediate_frame_helpers_load_demod_matrices(beam_num=beam)
-
-            for modstate in range(1, self.constants.num_modstates + 1):
-                # Lamp
-                ######
-                lamp_dict[VispTag.beam(beam)][
-                    VispTag.modstate(modstate)
-                ] = self.intermediate_frame_helpers_load_lamp_gain_array(
-                    beam=beam, modstate=modstate
+                demod_dict[VispTag.beam(beam)] = next(
+                    self.read(
+                        tags=[
+                            VispTag.intermediate_frame(beam=beam),
+                            VispTag.task_demodulation_matrices(),
+                        ],
+                        decoder=fits_array_decoder,
+                    )
                 )
 
+            for modstate in range(1, self.constants.num_modstates + 1):
                 # Solar
                 #######
-                solar_dict[VispTag.beam(beam)][
-                    VispTag.modstate(modstate)
-                ] = self.intermediate_frame_helpers_load_solar_gain_array(
-                    beam=beam, modstate=modstate
+                solar_dict[VispTag.beam(beam)][VispTag.modstate(modstate)] = next(
+                    self.read(
+                        tags=[
+                            VispTag.intermediate_frame(beam=beam, modstate=modstate),
+                            VispTag.task_solar_gain(),
+                        ],
+                        decoder=fits_array_decoder,
+                    )
                 )
 
                 # State Offset
                 ##############
-                state_offset_dict[VispTag.beam(beam)][
-                    VispTag.modstate(modstate)
-                ] = self.intermediate_frame_helpers_load_state_offset(beam=beam, modstate=modstate)
+                state_offset_dict[VispTag.beam(beam)][VispTag.modstate(modstate)] = next(
+                    self.read(
+                        tags=[
+                            VispTag.intermediate_frame(beam=beam, modstate=modstate),
+                            VispTag.task_geometric_offset(),
+                        ],
+                        decoder=fits_array_decoder,
+                    )
+                )
 
         return CalibrationCollection(
             dark=dark_dict,
             background=background_dict,
-            lamp_gain=lamp_dict,
             solar_gain=solar_dict,
             angle=angle_dict,
             state_offset=state_offset_dict,
@@ -239,7 +269,7 @@ class ScienceCalibration(
         - Apply telescope correction, if needed
         - Write calibrated arrays
         """
-        for exp_time in self.constants.observe_exposure_times:
+        for readout_exp_time in self.constants.observe_readout_exp_times:
             for map_scan in range(1, self.constants.num_map_scans + 1):
                 for raster_step in range(0, self.constants.num_raster_steps):
                     beam_storage = dict()
@@ -249,7 +279,7 @@ class ScienceCalibration(
                         with self.apm_processing_step(f"Basic corrections for {apm_str}"):
                             # Initialize array_stack and headers
                             if self.constants.correct_for_polarization:
-                                logging.info(
+                                logger.info(
                                     f"Processing polarimetric observe frames from {apm_str}"
                                 )
                                 (
@@ -259,11 +289,11 @@ class ScienceCalibration(
                                     beam=beam,
                                     raster_step=raster_step,
                                     map_scan=map_scan,
-                                    exp_time=exp_time,
+                                    readout_exp_time=readout_exp_time,
                                     calibrations=calibrations,
                                 )
                             else:
-                                logging.info(
+                                logger.info(
                                     f"Processing spectrographic observe frames from {apm_str}"
                                 )
                                 intermediate_array, intermediate_header = self.correct_single_frame(
@@ -271,25 +301,22 @@ class ScienceCalibration(
                                     modstate=1,
                                     raster_step=raster_step,
                                     map_scan=map_scan,
-                                    exp_time=exp_time,
+                                    readout_exp_time=readout_exp_time,
                                     calibrations=calibrations,
                                 )
-                                intermediate_header = self._compute_date_keys(intermediate_header)
+                                intermediate_header = self.compute_date_keys(intermediate_header)
                             beam_storage[VispTag.beam(beam)] = intermediate_array
                             header_storage[VispTag.beam(beam)] = intermediate_header
 
                     with self.apm_processing_step("Combining beams"):
-                        logging.info("Combining beams")
                         calibrated = self.combine_beams(beam_storage, header_storage, calibrations)
 
                     if self.constants.correct_for_polarization:
                         with self.apm_processing_step("Correcting telescope polarization"):
-                            logging.info("Correcting telescope polarization")
                             calibrated = self.telescope_polarization_correction(calibrated)
 
                     # Save the final output files
                     with self.apm_writing_step("Writing calibrated arrays"):
-                        logging.info("Writing calibrated arrays")
                         self.write_calibrated_array(
                             calibrated, map_scan=map_scan, calibrations=calibrations
                         )
@@ -299,7 +326,7 @@ class ScienceCalibration(
         beam: int,
         raster_step: int,
         map_scan: int,
-        exp_time: float,
+        readout_exp_time: float,
         calibrations: CalibrationCollection,
     ) -> tuple[np.ndarray, fits.Header]:
         """
@@ -308,7 +335,9 @@ class ScienceCalibration(
         This includes basic corrections and demodulation. Beam combination happens elsewhere.
         """
         # Create the 3D stack of corrected modulated arrays
-        array_shape = calibrations.dark[VispTag.beam(1)][VispTag.exposure_time(exp_time)].shape
+        array_shape = calibrations.dark[VispTag.beam(1)][
+            VispTag.readout_exp_time(readout_exp_time)
+        ].shape
         array_stack = np.zeros(array_shape + (self.constants.num_modstates,))
         header_stack = []
 
@@ -320,7 +349,7 @@ class ScienceCalibration(
                     modstate=modstate,
                     raster_step=raster_step,
                     map_scan=map_scan,
-                    exp_time=exp_time,
+                    readout_exp_time=readout_exp_time,
                     calibrations=calibrations,
                 )
                 # Add this result to the 3D stack
@@ -328,16 +357,19 @@ class ScienceCalibration(
                 header_stack.append(corrected_header)
 
         with self.apm_processing_step("Applying instrument polarization correction"):
-            logging.info("Applying instrument polarization correction")
-            intermediate_array = self.polarization_correction(
-                array_stack, calibrations.demod_matrices[VispTag.beam(beam)]
+            intermediate_array = nd_left_matrix_multiply(
+                vector_stack=array_stack,
+                matrix_stack=calibrations.demod_matrices[VispTag.beam(beam)],
             )
-            intermediate_header = self._compute_date_keys(header_stack)
+            intermediate_header = self.compute_date_keys(header_stack)
 
         return intermediate_array, intermediate_header
 
     def combine_beams(
-        self, array_dict: dict, header_dict: dict, calibrations: CalibrationCollection
+        self,
+        array_dict: dict[str, np.ndarray],
+        header_dict: dict[str, fits.Header],
+        calibrations: CalibrationCollection,
     ) -> VispL0FitsAccess:
         """
         Average all beams together and chop the resulting frame to just the region of overlap.
@@ -351,14 +383,13 @@ class ScienceCalibration(
             if fits.HeaderDiff(headers[0], h):
                 raise ValueError("Headers are different! This should NEVER happen!")
 
-        array_list = []
-        for beam in range(1, self.constants.num_beams + 1):
-            array_list.append(array_dict[VispTag.beam(beam)])
-
-        avg_array = average_numpy_arrays(array_list)
+        if self.constants.correct_for_polarization:
+            avg_array = self.combine_polarimetric_beams(array_dict)
+        else:
+            avg_array = self.combine_spectrographic_beams(array_dict)
 
         x_slice, y_slice = calibrations.beams_overlap_slice
-        logging.info(f"Trimming non-overlapping beam edges by ({x_slice}, {y_slice})")
+        logger.info(f"Trimming non-overlapping beam edges by ({x_slice}, {y_slice})")
         cut_array = avg_array[x_slice, y_slice]
 
         hdu = fits.ImageHDU(data=cut_array, header=headers[0])
@@ -366,6 +397,41 @@ class ScienceCalibration(
 
         return obj
 
+    def combine_polarimetric_beams(self, array_dict: dict[str, np.ndarray]) -> np.ndarray:
+        """
+        Combine polarimetric beams so that polarization states are normalized by the intensity state (Stokes I).
+
+        In other words:
+
+        avg_I = (beam1_I + beam2_I) / 2
+        avg_Q = (beam1_Q / beam1_I + beam2_Q / beam2_I) / 2. * avg_I
+
+        ...and the same for U and V
+        """
+        beam1_data = array_dict[VispTag.beam(1)]
+        beam2_data = array_dict[VispTag.beam(2)]
+
+        avg_data = np.zeros_like(beam1_data)
+        # Rely on the fact that the Stokes states are in order after demodulation
+        avg_I = (beam1_data[:, :, 0] + beam2_data[:, :, 0]) / 2.0
+        avg_data[:, :, 0] = avg_I
+
+        for stokes in range(1, 4):
+            beam1_norm = beam1_data[:, :, stokes] / beam1_data[:, :, 0]
+            beam2_norm = beam2_data[:, :, stokes] / beam2_data[:, :, 0]
+            avg_data[:, :, stokes] = avg_I * (beam1_norm + beam2_norm) / 2.0
+
+        return avg_data
+
+    def combine_spectrographic_beams(self, array_dict: dict[str, np.ndarray]) -> np.ndarray:
+        """Simply average the two beams together."""
+        array_list = []
+        for beam in range(1, self.constants.num_beams + 1):
+            array_list.append(array_dict[VispTag.beam(beam)])
+
+        avg_array = average_numpy_arrays(array_list)
+        return avg_array
+
     def write_calibrated_array(
         self,
         calibrated_object: VispL0FitsAccess,
@@ -385,25 +451,36 @@ class ScienceCalibration(
 
         map_scan
             The current map scan. Needed because it's not a header key
+
+        calibrations
+            Calibration collection
+
+        Returns
+        -------
+        None
+
         """
         # We only need to compute the header once
         #  (In fact we *need* to only compute it once because we update WCS values in place and running this function
         #   more than once would result in incorrect WCS info).
-        final_header = self._update_calibrated_header(
+        final_header = self.update_calibrated_header(
             calibrated_object.header, map_scan=map_scan, calibrations=calibrations
         )
         if self.constants.correct_for_polarization:  # Write all 4 stokes params
+            stokes_I_data = calibrated_object.data[:, :, 0]
             for i, stokes_param in enumerate(self.constants.stokes_params):
-                final_data = self._re_dummy_data(calibrated_object.data[:, :, i])
+                stokes_data = calibrated_object.data[:, :, i]
+                final_data = self.re_dummy_data(stokes_data)
+                pol_header = self.add_L1_pol_headers(final_header, stokes_data, stokes_I_data)
                 self.write_cal_array(
                     data=final_data,
-                    header=final_header,
+                    header=pol_header,
                     stokes=stokes_param,
                     raster_step=calibrated_object.raster_scan_step,
                     map_scan=map_scan,
                 )
         else:  # Only write stokes I
-            final_data = self._re_dummy_data(calibrated_object.data)
+            final_data = self.re_dummy_data(calibrated_object.data)
             self.write_cal_array(
                 data=final_data,
                 header=final_header,
@@ -418,7 +495,7 @@ class ScienceCalibration(
         modstate: int,
         raster_step: int,
         map_scan: int,
-        exp_time: float,
+        readout_exp_time: float,
         calibrations: CalibrationCollection,
     ) -> tuple[np.ndarray, fits.Header]:
         """
@@ -435,48 +512,62 @@ class ScienceCalibration(
         ----------
         beam
             The beam number for this single step
+
         modstate
             The modulator state for this single step
+
         raster_step
             The slit step for this single step
+
         map_scan
             The current map scan
-        exp_time
+
+        readout_exp_time
             The exposure time for this single step
+
         calibrations
             Collection of all calibration objects
 
+
         Returns
         -------
+        tuple[np.ndarray, fits.Header]
             Corrected array, header
         """
         # Extract calibrations
-        dark_array = calibrations.dark[VispTag.beam(beam)][VispTag.exposure_time(exp_time)]
+        dark_array = calibrations.dark[VispTag.beam(beam)][
+            VispTag.readout_exp_time(readout_exp_time)
+        ]
         background_array = calibrations.background[VispTag.beam(beam)]
-        lamp_gain_array = calibrations.lamp_gain[VispTag.beam(beam)][VispTag.modstate(modstate)]
         solar_gain_array = calibrations.solar_gain[VispTag.beam(beam)][VispTag.modstate(modstate)]
         angle = calibrations.angle[VispTag.beam(beam)]
         spec_shift = calibrations.spec_shift[VispTag.beam(beam)]
         state_offset = calibrations.state_offset[VispTag.beam(beam)][VispTag.modstate(modstate)]
 
         # Grab the input observe frame
+        tags = [
+            VispTag.input(),
+            VispTag.frame(),
+            VispTag.task_observe(),
+            VispTag.modstate(modstate),
+            VispTag.map_scan(map_scan),
+            VispTag.raster_step(raster_step),
+            VispTag.readout_exp_time(readout_exp_time),
+        ]
         observe_object_list = list(
-            self.input_frame_loaders_observe_fits_access_generator(
-                map_scan=map_scan,
-                raster_step=raster_step,
-                modstate=modstate,
-                exposure_time=exp_time,
-            )
+            self.read(tags=tags, decoder=fits_access_decoder, fits_access_class=VispL0FitsAccess)
         )
+
         if len(observe_object_list) > 1:
             raise ValueError(
                 f"Found more than one observe frame for {map_scan = }, {raster_step = }, {modstate = }, "
-                f"and {exp_time = }. This should NEVER have happened!"
+                f"and {readout_exp_time = }. This should NEVER have happened!"
             )
         observe_object = observe_object_list[0]
 
         # Split the beam we want
-        observe_data = self.input_frame_loaders_get_beam(observe_object.data, beam=beam)
+        readout_normalized_data = observe_object.data / observe_object.num_raw_frames_per_fpa
+        observe_data = self.beam_access_get_beam(readout_normalized_data, beam=beam)
 
         # Dark correction
         dark_corrected_array = next(subtract_array_from_arrays(observe_data, dark_array))
@@ -486,13 +577,10 @@ class ScienceCalibration(
             subtract_array_from_arrays(dark_corrected_array, background_array)
         )
 
-        # Lamp gain correction
-        lamp_corrected_array = next(
-            divide_arrays_by_array(background_corrected_array, lamp_gain_array)
-        )
-
         # Solar gain correction
-        solar_corrected_array = next(divide_arrays_by_array(lamp_corrected_array, solar_gain_array))
+        solar_corrected_array = next(
+            divide_arrays_by_array(background_corrected_array, solar_gain_array)
+        )
 
         # Geo correction
         geo_corrected_array = next(
@@ -506,28 +594,6 @@ class ScienceCalibration(
 
         return spectral_corrected_array, observe_object.header
 
-    @staticmethod
-    def polarization_correction(array_stack: np.ndarray, demod_matrices: np.ndarray) -> np.ndarray:
-        """
-        Apply a polarization correction to an array by multiplying the array stack by the demod matrices.
-
-        Parameters
-        ----------
-        array_stack : np.ndarray
-            (x, y, M) stack of corrected arrays with M modulation states
-
-        demod_matrices : np.ndarray
-            (x, y, 4, M) stack of demodulation matrices with 4 stokes planes and M modulation states
-
-
-        Returns
-        -------
-        np.ndarray
-            (x, y, 4) ndarray with the planes being IQUV
-        """
-        demodulated_array = np.sum(demod_matrices * array_stack[:, :, None, :], axis=3)
-        return demodulated_array
-
     def telescope_polarization_correction(
         self,
         inst_demod_obj: VispL0FitsAccess,
@@ -539,20 +605,23 @@ class ScienceCalibration(
         ----------
         inst_demod_obj
             A demodulated, beam averaged frame
-        telescope_db : str
-            Telescope polarization correction loaded from telescope database
 
         Returns
         -------
         FitsAccess object with telescope corrections applied
+
         """
         tm = Telescope.from_fits_access(inst_demod_obj)
-        mueller_matrix = tm.generate_inverse_telescope_model(M12=True, include_parallactic=True)
-        inst_demod_obj.data = self.polarization_correction(inst_demod_obj.data, mueller_matrix)
+        mueller_matrix = tm.generate_inverse_telescope_model(
+            M12=True, rotate_to_fixed_SDO_HINODE_polarized_frame=True, swap_UV_signs=True
+        )
+        inst_demod_obj.data = nd_left_matrix_multiply(
+            vector_stack=inst_demod_obj.data, matrix_stack=mueller_matrix
+        )
         return inst_demod_obj
 
     @staticmethod
-    def _compute_date_keys(headers: Iterable[fits.Header] | fits.Header) -> fits.Header:
+    def compute_date_keys(headers: Iterable[fits.Header] | fits.Header) -> fits.Header:
         """
         Generate correct DATE-??? header keys from a set of input headers.
 
@@ -560,6 +629,11 @@ class ScienceCalibration(
         * DATE-BEG - The (Spec-0122) DATE-OBS of the earliest input header
         * DATE-END - The (Spec-0122) DATE-OBS of the latest input header, plus the FPA exposure time
 
+        Parameters
+        ----------
+        headers : Iterable[fits.Header] | fits.Header
+            Headers
+
         Returns
         -------
         fits.Header
@@ -583,7 +657,7 @@ class ScienceCalibration(
 
         return header
 
-    def _re_dummy_data(self, data: np.ndarray):
+    def re_dummy_data(self, data: np.ndarray):
         """
         Add the dummy dimension that we have been secretly squeezing out during processing.
 
@@ -593,11 +667,15 @@ class ScienceCalibration(
         ----------
         data : np.ndarray
             Corrected data
+
+        Returns
+        -------
+        None
+
         """
-        logging.debug(f"Adding dummy WCS dimension to array with shape {data.shape}")
         return data[None, :, :]
 
-    def _update_calibrated_header(
+    def update_calibrated_header(
         self, header: fits.Header, map_scan: int, calibrations: CalibrationCollection
     ) -> fits.Header:
         """
@@ -607,15 +685,17 @@ class ScienceCalibration(
 
         Parameters
         ----------
-        header
+        header : fits.Header
             The header to update
 
-        map_scan
+        map_scan : int
             Current map scan
 
         Returns
         -------
         fits.Header
+            Updated header
+
         """
         # Update the map scan number
         header["VSPNMAPS"] = self.constants.num_map_scans
@@ -627,27 +707,73 @@ class ScienceCalibration(
         # This if catches 0's and Nones
         if x_slice.start:
             # .start will only be non-None or 0 if the slice is from the start. In this case we need to update the WCS
-            logging.info(
+            logger.info(
                 f"Adjusting CRPIX2 from {header['CRPIX2']} to {header['CRPIX2'] - x_slice.start}"
             )
             header["CRPIX2"] = header["CRPIX2"] - x_slice.start
-            logging.info(
+            logger.info(
                 f"Adjusting CRPIX2A from {header['CRPIX2A']} to {header['CRPIX2A'] - x_slice.start}"
             )
             header["CRPIX2A"] = header["CRPIX2A"] - x_slice.start
 
         if y_slice.start:
-            logging.info(
+            logger.info(
                 f"Adjusting CRPIX1 from {header['CRPIX1']} to {header['CRPIX1'] - y_slice.start}"
             )
             header["CRPIX1"] = header["CRPIX1"] - y_slice.start
-            logging.info(
+            logger.info(
                 f"Adjusting CRPIX1A from {header['CRPIX1A']} to {header['CRPIX1A'] - y_slice.start}"
             )
             header["CRPIX1A"] = header["CRPIX1A"] - y_slice.start
 
         return header
 
+    def add_L1_pol_headers(
+        self, input_header: fits.Header, stokes_data: np.ndarray, stokes_I_data: np.ndarray
+    ) -> fits.Header:
+        """Compute and add 214 header values specific to polarimetric datasets."""
+        # Probably not needed, but just to be safe
+        output_header = input_header.copy()
+
+        pol_noise = self.compute_polarimetric_noise(stokes_data, stokes_I_data)
+        pol_sensitivity = self.compute_polarimetric_sensitivity(stokes_I_data)
+        output_header["POL_NOIS"] = pol_noise
+        output_header["POL_SENS"] = pol_sensitivity
+
+        return output_header
+
+    def compute_polarimetric_noise(
+        self, stokes_data: np.ndarray, stokes_I_data: np.ndarray
+    ) -> float:
+        r"""
+        Compute the polarimetric noise for a single frame.
+
+        The polarimetric noise, :math:`N`, is defined as
+
+        .. math::
+
+            N = stddev(\frac{F_i}{F_I})
+
+        where :math:`F_i` is a full array of values for Stokes parameter :math:`i` (I, Q, U, V), and :math:`F_I` is the
+        full frame of Stokes-I. The stddev is computed across the entire frame.
+        """
+        return float(np.nanstd(stokes_data / stokes_I_data))
+
+    def compute_polarimetric_sensitivity(self, stokes_I_data: np.ndarray) -> float:
+        r"""
+        Compute the polarimetric sensitivity for a single frame.
+
+        The polarimetric sensitivity is the smallest signal that can be measured based on the values in the Stokes-I
+        frame. The sensitivity, :math:`S`, is computed as
+
+        .. math::
+
+            S = \frac{1}{\sqrt{\mathrm{max}(F_I)}}
+
+        where :math:`F_I` is the full frame of values for Stokes-I.
+        """
+        return float(1.0 / np.sqrt(np.nanmax(stokes_I_data)))
+
     def write_cal_array(
         self,
         data: np.ndarray,
@@ -680,6 +806,7 @@ class ScienceCalibration(
         Returns
         -------
         None
+
         """
         tags = [
             VispTag.calibrated(),
@@ -689,7 +816,7 @@ class ScienceCalibration(
             VispTag.map_scan(map_scan),
         ]
         hdul = fits.HDUList([fits.PrimaryHDU(), fits.CompImageHDU(header=header, data=data)])
-        self.fits_data_write(hdu_list=hdul, tags=tags)
+        self.write(data=hdul, tags=tags, encoder=fits_hdulist_encoder)
 
         filename = next(self.read(tags=tags))
-        logging.info(f"Wrote intermediate file for {tags = } to {filename}")
+        logger.info(f"Wrote intermediate file for {tags = } to {filename}")
