@@ -6,9 +6,9 @@ import numpy as np
 import pytest
 from astropy.io import fits
 from astropy.time import Time
-from astropy.time import TimeDelta
 from dkist_header_validator.translator import translate_spec122_to_spec214_l0
 from dkist_processing_common._util.scratch import WorkflowFileSystem
+from dkist_processing_common.codecs.fits import fits_hdulist_encoder
 from dkist_processing_common.parsers.single_value_single_key_flower import (
     SingleValueSingleKeyFlower,
 )
@@ -23,8 +23,8 @@ from dkist_processing_visp.parsers.raster_step import RasterScanStepFlower
 from dkist_processing_visp.parsers.visp_l0_fits_access import VispL0FitsAccess
 from dkist_processing_visp.tasks.parse import ParseL0VispInputData
 from dkist_processing_visp.tasks.parse import S
-from dkist_processing_visp.tests.conftest import VispHeadersValidObserveFrames
-from dkist_processing_visp.tests.conftest import VispTestingParameters
+from dkist_processing_visp.tests.conftest import VispInputDatasetParameterValues
+from dkist_processing_visp.tests.header_models import VispHeadersValidObserveFrames
 
 
 @pytest.fixture(scope="session")
@@ -32,61 +32,14 @@ def complete_map_headers():
     num_steps = 5
     num_modstates = 4
     num_maps = 3
-    start_time = Time("1946-11-20T12:00:00")
-    dt = TimeDelta(10, format="sec")
-    i = 0
-    header_list = []
-    map_number_list = []
-    for map_scan in range(1, num_maps + 1):
-        for step_num in range(0, num_steps):
-            for modstate in range(1, num_modstates + 1):
-                ds = VispHeadersValidObserveFrames(
-                    dataset_shape=(1, 2, 2),
-                    array_shape=(1, 2, 2),
-                    time_delta=10,
-                    num_raster_steps=num_steps,
-                    raster_step=step_num,
-                    num_modstates=num_modstates,
-                    modstate=modstate,
-                    start_time=(start_time + i * dt).to_datetime(),
-                )
-                header = translate_spec122_to_spec214_l0(next(ds).header())
-                header_list.append(header)
-                map_number_list.append(map_scan)
-                i += 1
-
-    return header_list, num_steps, num_modstates, num_maps
-
-
-@pytest.fixture(scope="session")
-def map_headers_with_missing_steps():
-    num_steps = 2
-    num_modstates = 2
-    num_maps = 3
-    start_time = Time("1946-11-20T12:00:00")
-    dt = TimeDelta(10, format="sec")
-    i = 0
-    header_list = []
-    map_number_list = []
-    for map_scan in range(1, num_maps + 1):
-        for step_num in range(0, num_steps):
-            if step_num == 0 and map_scan == 1:
-                continue
-            for modstate in range(1, num_modstates + 1):
-                ds = VispHeadersValidObserveFrames(
-                    dataset_shape=(1, 2, 2),
-                    array_shape=(1, 2, 2),
-                    time_delta=10,
-                    num_raster_steps=num_steps,
-                    raster_step=step_num,
-                    num_modstates=num_modstates,
-                    modstate=modstate,
-                    start_time=(start_time + i * dt).to_datetime(),
-                )
-                header = translate_spec122_to_spec214_l0(next(ds).header())
-                header_list.append(header)
-                map_number_list.append(map_scan)
-                i += 1
+    dataset = VispHeadersValidObserveFrames(
+        array_shape=(1, 2, 2),
+        time_delta=10,
+        num_maps=num_maps,
+        num_raster_steps=num_steps,
+        num_modstates=num_modstates,
+    )
+    header_list = [translate_spec122_to_spec214_l0(d.header()) for d in dataset]
 
     return header_list, num_steps, num_modstates, num_maps
 
@@ -102,9 +55,33 @@ def organized_fits_access_dict(complete_map_headers):
     return all_dict
 
 
+def write_obs_frames_with_multiple_exp_per_step_to_task(task, num_exp: int):
+    array_shape = (1, 2, 2)
+    num_maps = 2
+    num_steps = 3
+    num_modstates = 1
+    dataset = VispHeadersValidObserveFrames(
+        array_shape=array_shape,
+        time_delta=10,
+        num_maps=num_maps,
+        num_raster_steps=num_steps,
+        num_modstates=num_modstates,
+    )
+    header_list = [translate_spec122_to_spec214_l0(d.header()) for d in dataset]
+    for header in header_list:
+        hdu = fits.PrimaryHDU(data=np.ones(array_shape), header=fits.Header(header))
+        hdul = fits.HDUList([hdu])
+
+        # Make multiple exposures by just writing the same header twice
+        for _ in range(num_exp):
+            task.write(
+                data=hdul, tags=[VispTag.input(), VispTag.frame()], encoder=fits_hdulist_encoder
+            )
+
+
 class ParseTaskJustMapStuff(ParseL0VispInputData):
     @property
-    def constant_flowers(self) -> list[S]:
+    def constant_buds(self) -> list[S]:
         return [NumMapScansBud()]
 
     @property
@@ -119,110 +96,24 @@ class ParseTaskJustMapStuff(ParseL0VispInputData):
 
 
 @pytest.fixture()
-def map_only_parse_task_with_correct_map(
-    complete_map_headers, tmp_path_factory, recipe_run_id, assign_input_dataset_doc_to_task
-):
+def map_only_parse_task(tmp_path_factory, recipe_run_id, assign_input_dataset_doc_to_task):
 
-    complete_headers, num_steps, num_modstates, num_maps = complete_map_headers
     with ParseTaskJustMapStuff(
         recipe_run_id=recipe_run_id, workflow_name="parse_map_scans", workflow_version="X.Y.Z"
     ) as task:
         try:
-            task._scratch = WorkflowFileSystem(
+            task.scratch = WorkflowFileSystem(
                 scratch_base_path=tmp_path_factory.mktemp("map_scan"), recipe_run_id=recipe_run_id
             )
-            assign_input_dataset_doc_to_task(task, VispTestingParameters())
-            for header in complete_headers:
-                hdu = fits.PrimaryHDU(data=np.ones((1, 2, 2)), header=fits.Header(header))
-                hdul = fits.HDUList([hdu])
-                task.fits_data_write(hdu_list=hdul, tags=[VispTag.input(), VispTag.frame()])
+            assign_input_dataset_doc_to_task(task, VispInputDatasetParameterValues())
+
+            yield task
 
-            yield task, num_steps, num_modstates, num_maps
         except:
             raise
+
         finally:
-            task.scratch.purge()
-            task.constants._purge()
-
-
-@pytest.fixture()
-def map_only_parse_task_with_missing_steps(
-    map_headers_with_missing_steps,
-    tmp_path_factory,
-    recipe_run_id,
-    assign_input_dataset_doc_to_task,
-):
-
-    complete_headers, num_steps, num_modstates, num_maps = map_headers_with_missing_steps
-    with ParseTaskJustMapStuff(
-        recipe_run_id=recipe_run_id, workflow_name="parse_map_scans", workflow_version="X.Y.Z"
-    ) as task:
-        try:
-            task._scratch = WorkflowFileSystem(
-                scratch_base_path=tmp_path_factory.mktemp("map_scan"), recipe_run_id=recipe_run_id
-            )
-            assign_input_dataset_doc_to_task(task, VispTestingParameters())
-            for header in complete_headers:
-                hdu = fits.PrimaryHDU(data=np.ones((1, 2, 2)), header=fits.Header(header))
-                hdul = fits.HDUList([hdu])
-                task.fits_data_write(hdu_list=hdul, tags=[VispTag.input(), VispTag.frame()])
-
-            yield task, num_steps, num_modstates, num_maps
-        except:
-            raise
-        finally:
-            task.scratch.purge()
-            task.constants._purge()
-
-
-@pytest.fixture
-def map_only_parse_task_with_multiple_exposures_per_raster_step(
-    tmp_path, recipe_run_id, assign_input_dataset_doc_to_task
-):
-    num_steps = 4
-    num_exp = 3
-    num_modstates = 2
-    start_time = Time("1946-11-20T12:00:00")
-    dt = TimeDelta(10, format="sec")
-    i = 0
-    with ParseTaskJustMapStuff(
-        recipe_run_id=recipe_run_id,
-        workflow_name="parse_visp_input_data",
-        workflow_version="VX.Y",
-    ) as task:
-        try:  # This try... block is here to make sure the dbs get cleaned up if there's a failure in the fixture
-            task._scratch = WorkflowFileSystem(
-                scratch_base_path=tmp_path, recipe_run_id=recipe_run_id
-            )
-            assign_input_dataset_doc_to_task(task, VispTestingParameters())
-            for s in range(num_steps):
-                for m in range(1, num_modstates + 1):
-                    ds = VispHeadersValidObserveFrames(
-                        dataset_shape=(num_exp, 2, 2),
-                        array_shape=(1, 2, 2),
-                        time_delta=0,  # This is where we produce multiple exposures
-                        num_raster_steps=num_steps,
-                        raster_step=s,
-                        num_modstates=num_modstates,
-                        modstate=m,
-                        polarimeter_mode="observe_polarimetric",
-                        start_time=(start_time + i * dt).to_datetime(),
-                    )
-                    for d in ds:
-                        header = d.header()
-                        translated_header = translate_spec122_to_spec214_l0(header)
-                        hdu = fits.PrimaryHDU(
-                            data=np.ones((1, 2, 2)), header=fits.Header(translated_header)
-                        )
-                        hdul = fits.HDUList([hdu])
-                        task.fits_data_write(hdu_list=hdul, tags=[VispTag.input(), VispTag.frame()])
-                    i += 1
-            yield task, num_exp
-        except:
-            raise
-        finally:
-            task.scratch.purge()
-            task.constants._purge()
+            task._purge()
 
 
 def test_singlescanstep_correct(organized_fits_access_dict):
@@ -248,13 +139,20 @@ def test_singlescanstep_correct(organized_fits_access_dict):
     assert len(set(flat_obj_list)) == len(flat_obj_list)
 
 
-def test_parse_map_repeats(map_only_parse_task_with_correct_map):
+def test_parse_map_repeats(map_only_parse_task, complete_map_headers):
     """
     :Given: A map-parsing task with files that represent complete maps
     :When: Parsing the files
     :Then: The correct number of map scans is inferred and each file is tagged correctly
     """
-    task, num_steps, num_modstates, num_maps = map_only_parse_task_with_correct_map
+    task = map_only_parse_task
+    complete_headers, num_steps, num_modstates, num_maps = complete_map_headers
+
+    for header in complete_headers:
+        hdu = fits.PrimaryHDU(data=np.ones((1, 2, 2)), header=fits.Header(header))
+        hdul = fits.HDUList([hdu])
+        task.write(data=hdul, tags=[VispTag.input(), VispTag.frame()], encoder=fits_hdulist_encoder)
+
     task()
 
     assert task.constants._db_dict[VispBudName.num_map_scans.value] == num_maps
@@ -286,14 +184,16 @@ def test_parse_map_repeats(map_only_parse_task_with_correct_map):
 
 
 def test_multiple_exp_per_step_raises_error(
-    map_only_parse_task_with_multiple_exposures_per_raster_step,
+    map_only_parse_task,
 ):
     """
     :Given: A map-parsing task with data that has multiple exposures per (raster, modstate, map_scan)
     :When: Calling the parse task
     :Then: The correct error is raised
     """
-    task, num_exp = map_only_parse_task_with_multiple_exposures_per_raster_step
+    task = map_only_parse_task
+    num_exp = 3
+    write_obs_frames_with_multiple_exp_per_step_to_task(task=task, num_exp=num_exp)
     with pytest.raises(
         ValueError,
         match=re.escape(
